package jdk.management.resource;

import java.util.Collections;
import java.util.stream.Stream;
import jdk.management.resource.internal.TotalResourceContext;
import jdk.management.resource.internal.SimpleResourceContext;
import java.security.AccessController;
import jdk.management.resource.internal.WrapInstrumentation;
import java.security.SecureRandom;
import java.util.concurrent.ThreadLocalRandom;
import jdk.management.resource.internal.UnassignedContext;
import jdk.management.resource.internal.ResourceNatives;
import java.util.Set;

public final class ResourceContextFactory
{
    private static final ResourceContextFactory instance;
    private final ResourceContext unassigned;
    private static Set<ResourceType> supportedResourceTypes;
    private volatile boolean initialized;
    
    private ResourceContextFactory() {
        this.initialized = false;
        this.unassigned = (ResourceNatives.isEnabled() ? UnassignedContext.getUnassignedContext() : null);
    }
    
    public static boolean isEnabled() {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(new RuntimePermission("jdk.management.resource.getResourceContextFactory"));
        }
        return ResourceNatives.isEnabled();
    }
    
    public static ResourceContextFactory getInstance() {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(new RuntimePermission("jdk.management.resource.getResourceContextFactory"));
        }
        if (!ResourceNatives.isEnabled()) {
            throw new UnsupportedOperationException("Resource management is not enabled");
        }
        ResourceContextFactory.instance.initInstrumentation();
        return ResourceContextFactory.instance;
    }
    
    private synchronized void initInstrumentation() throws InternalError {
        if (!this.initialized) {
            ThreadLocalRandom.current();
            new SecureRandom();
            try {
                ((Runnable)Class.forName("jdk.management.resource.internal.inst.InitInstrumentation").newInstance()).run();
            }
            catch (ClassNotFoundException ex) {}
            catch (IllegalAccessException | InstantiationException cause) {
                throw new InternalError("Resource management instrumentation failed", cause);
            }
            if (!new WrapInstrumentation().wrapComplete()) {
                throw new InternalError("Resource management instrumentation failed");
            }
            this.initPreBoundThreads();
            this.initialized = true;
        }
    }
    
    private void initPreBoundThreads() {
        AccessController.doPrivileged(()/*  */ -> {
            ThreadGroup parent = Thread.currentThread().getThreadGroup();
            while (parent.getParent() != null) {
                parent = parent.getParent();
            }
            Thread[] list = new Thread[parent.activeCount() * 2];
            int i = parent.enumerate(list, true);
            int n;
            while (i < n) {
                if (list[i] != null) {
                    (list[i].getThreadGroup().equals(parent) ? UnassignedContext.getSystemContext() : UnassignedContext.getUnassignedContext()).bindThreadContext(list[i]);
                }
                ++i;
            }
            return null;
        });
    }
    
    public ResourceContext create(String s) {
        return SimpleResourceContext.create(s);
    }
    
    public ResourceContext lookup(String s) {
        return SimpleResourceContext.get(s);
    }
    
    public ResourceContext getThreadContext() {
        return this.getThreadContext(Thread.currentThread());
    }
    
    public ResourceContext getThreadContext(Thread thread) {
        return SimpleResourceContext.getThreadContext(thread);
    }
    
    public ResourceRequest getResourceRequest(ResourceType resourceType) {
        return this.getThreadContext().getResourceRequest(resourceType);
    }
    
    public ResourceContext getUnassignedContext() {
        return this.unassigned;
    }
    
    public ResourceContext getTotalsContext() {
        return TotalResourceContext.getTotalContext();
    }
    
    public Stream<ResourceContext> contexts() {
        return SimpleResourceContext.contexts();
    }
    
    public Set<ResourceType> supportedResourceTypes() {
        synchronized (this) {
            if (ResourceContextFactory.supportedResourceTypes == null) {
                Set<ResourceType> builtinTypes = ResourceType.builtinTypes();
                if (!ResourceNatives.isHeapRetainedEnabled()) {
                    builtinTypes.remove(ResourceType.HEAP_RETAINED);
                }
                ResourceContextFactory.supportedResourceTypes = Collections.unmodifiableSet(builtinTypes);
            }
            return ResourceContextFactory.supportedResourceTypes;
        }
    }
    
    static {
        instance = new ResourceContextFactory();
        ResourceContextFactory.supportedResourceTypes = null;
    }
}
